/* jshint undef: true, esnext: true */
/* global plugin, require, Packages, java */

var _ = require("lodash");
var on = require("on")(plugin);
var Objects = require("convenience").Objects;
var facade = require("facade");
var request = require("request");
var OSSystemState = Packages.cattura.utils.os.OSSystemState;
var StringUtilities = Packages.cattura.utils.string.StringUtilities;
var EventType = Packages.cattura.agents.configuration.model.vos.events.EventType;
var STORE = plugin.info().dataStore;
var settings = plugin.info().settings;
var LOGGER = plugin.info().logger;
var Level = java.util.logging.Level;

var EventHelper = Packages.cattura.agents.configuration.model.utilities.EventHelper;
var StringTemplateParser = Packages.jodd.template.StringTemplateParser;
var BeanUtil = Packages.jodd.bean.BeanUtil;

var notifiers = _(Objects.toJS(STORE.select("notifier:*")))
    .map(notifier => [notifier.id, new Notifier(notifier)])
    .fromPairs()
    .value();

/*
 * By default this plugin will handle anything that's not an internal event,
 * or is on the allowlist, or is not on the blocklist.
 */
var EVENT_TYPE_ALLOWLIST = [];
var EVENT_TYPE_BLOCKLIST = ["AudioLevelsEvent"];

function eventIsNotifiable (event)
{
    var helper = new EventHelper(event);
    return (
        !!_(EVENT_TYPE_ALLOWLIST).find(eventType => helper.isOfType(eventType)) ||
        !(
            helper.isInternalEvent() || 
            !!_(EVENT_TYPE_BLOCKLIST).find(eventType => helper.eventIsOfType(eventType))
        )
    );
}

function makeDefaultRequestBodyForEvent (event)
{
    return Objects.toJS({
        eventSource: getUnitName(),
        eventType: event.type,
        eventTime: event.time.time,
        eventDescription: event.description,
        eventData: event.data
    }, true);
}

function interpolateEventData (event, target)
{
    target = Objects.toJS(target, true);
    
    if (typeof target === "string")
    {
        return new StringTemplateParser().parse(target, macro => BeanUtil.pojo.getProperty(event, macro)) + "";
    }
    
    if (typeof target === "object")
    {
        if (_(target).isArray())
        {
            return _(target)
                .map(value => interpolateEventData(event, value))
                .value();
        }
        
        return _(target)
            .mapValues(value => interpolateEventData(event, value))
            .value();
    }
    
    return target;
}

function Notifier (data)
{
    _(this).extend(Objects.toJS(data, true)).value();
    
    this.canHandleEvent = function canHandleEvent (event)
    {
        if (!eventIsNotifiable(event))
        {
            return false;
        }

        return _(this.events).find(function (type) {
            return type == event.eventType.name() || type == event.getClass().simpleName;
        });
    };
    
    this.handleEvent = function handleEvent (event)
    {
        request.post(this.url + "", this.toRequestData(event), function (err, response, body) {
            if (err)
            {
                if (err.javaException)
                {
                    err.javaException.printStackTrace();
                }
                
                LOGGER.warning("Error: " + JSON.stringify(err));
            }
        });
    };
    
    this.toRequestData = function toRequestData (event)
    {
        let requestData = {};
        
        if (this.headers)
        {
            requestData.headers = interpolateEventData(event, this.headers);
        }
        
        if (this.contentType)
        {
            requestData.headers = requestData.headers || {};
            requestData.headers['Content-Type'] = this.contentType;
        }
        
        let body = this.payload ? interpolateEventData(event, this.payload)
                                : makeDefaultRequestBodyForEvent(event);
        
        if (this.isJson() || typeof body === "string")
        {
            if (typeof body !== "string")
            {
                body = JSON.stringify(body);
            }
            
            requestData.body = body;
        }
        else
        {
            requestData.form = _(body)
                .mapValues(value => typeof value === "object" ? JSON.stringify(value)
                                                              : value)
                .value();
        }
        
        return Objects.toJS(requestData, true);
    };
    
    this.isJson = function isJson ()
    {
        let contentType = this.contentType || (this.headers && this.headers['Content-Type'] ? this.headers['Content-Type']
                                                                                            : null);
        return contentType == "application/json";
    };
    
    this.isForm = function isForm ()
    {
        return !this.isJson();
    };
}

function ensureValidURL (url)
{
    if (!(_(url).startsWith("http://") || _(url).startsWith("https://")))
    {
        return "http://" + url;
    }
    
    return url;
}

function testNotifier (url, events, contentType, payload, event)
{
    new Notifier({
        url: ensureValidURL(url),
        events: events,
        contentType: contentType,
        payload: payload
    }).handleEvent(event);
}

function addNotifier (id, url, events, contentType, payload)
{
    var notifier = {
        id: id,
        url: ensureValidURL(url),
        events: events,
        contentType: contentType,
        payload: payload
    };
    
    STORE.set("notifier:" + id, notifier);
    
    notifier = new Notifier(notifier);
    notifiers[id] = notifier;
    return notifier;
}

function removeNotifier (id)
{
    delete notifiers[id];
    STORE.remove("notifier:" + id);
}

function getNotifiersForEvent (event)
{
    if (!eventIsNotifiable(event))
    {
        return [];
    }
    
    return _(notifiers)
        .filter(notifier => notifier.canHandleEvent(event))
        .value();
}

function generateIdForNotifier ()
{
    var id = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
    if (_(notifiers).find(notifier => notifier.id == id))
    {
        return generateIdForNotifier();
    }

    return id;
}

function getUnitName ()
{
    var settings = facade.retrieveProxy("SettingsProxy");
    var unitName = settings.getUserSettings().getString("unit-name").orElse("");
    if (!unitName.length())
    {
        unitName = OSSystemState.getMachineName();
    }
    return unitName;
}

function getExceptionFromEvent (event)
{
    if ('incident' in event)
    {
        return event.incident.exception;
    }
    else if ('exception' in event)
    {
        return event.exception;
    }
}

/*
 * Add a webhook with a custom payload.
 */
plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "url", description: "URL", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "contentType", description: "Content type of request", type: "java.lang.String"},
        {name: "payload", description: "Request body as JSON string", type: "java.lang.String"}
    ],
    handler: function (url, events, contentType, payload) {
        return addNotifier(generateIdForNotifier(), url, events, contentType, payload);
    }
});

plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"},
        {name: "url", description: "URL", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "contentType", description: "Content type of request", type: "java.lang.String"},
        {name: "payload", description: "Request body as JSON string", type: "java.lang.String"}
    ],
    handler: addNotifier
});

plugin.service({
    name: "testNotifier",
    description: "Test a notifier with a test event.",
    parameters: [
        {name: "url", description: "URL", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "contentType", description: "Content type of request", type: "java.lang.String"},
        {name: "payload", description: "Request body as JSON string", type: "java.lang.String"},
        {name: "event", description: "Test event", type: "java.lang.Object"}
    ],
    handler: testNotifier
});

/*
 * Add a webhook with a default payload.
 */
plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "url", description: "URL", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "contentType", description: "Content type of request", type: "java.lang.String"}
    ],
    handler: function (url, events, contentType, payload) {
        return addNotifier(generateIdForNotifier(), url, events, contentType, payload);
    }
});

plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"},
        {name: "url", description: "URL", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "contentType", description: "Content type of request", type: "java.lang.String"}
    ],
    handler: addNotifier
});

plugin.service({
    name: "testNotifier",
    description: "Test a notifier with a test event.",
    parameters: [
        {name: "url", description: "URL", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "contentType", description: "Content type of request", type: "java.lang.String"},
        {name: "event", description: "Test event", type: "java.lang.Object"}
    ],
    handler: testNotifier
});

plugin.service({
    name: "removeNotifier",
    description: "Remove a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"}
    ],
    handler: removeNotifier
});

plugin.service({
    name: "getNotifiers",
    description: "Return all registered notifiers",
    parameters: [],
    handler: () => _(notifiers).values().value()
});

function handleExceptionFromHandler (event, exception)
{
    var unwrappedException = exception.javaException || exception.rhinoException || exception;
    if (getExceptionFromEvent(event))
    {
        LOGGER.log(Level.WARNING, "Webhook Plugin encountered (and swallowed) exception while trying to handle exception event", unwrappedException);
        return;
    }

    facade.sendNotification("HANDLE_EXCEPTION", new java.lang.RuntimeException("Webhook Plugin encountered exception while trying to handle event: " + event, unwrappedException));
}

plugin.onEvent(function(event) {
    try
    {
        _(getNotifiersForEvent(event)).each(function (notifier) {
            notifier.handleEvent(event);
        });
    }
    catch (exception)
    {
        handleExceptionFromHandler(event, exception);
    }
});
